
#include <unistd.h> 
#include "config_and_coord.h"
#include "org_and_env.h"
#include <time.h>
#include "phenotypes.h" 


int main(int argc, char*argv[]){

  world*earth;
  int i, size, generation;
  population_information* info;

  int opt;
  static const char *optstring = "x:s:b:r:d:w:p:h?";

  struct{
    int steps;
    int steps2bmp;
    int resolution;
    bool specialBmpDir;
    char bmpDir[16];
    int worldSizeGlobal;
    int popFactor;
    int popSizeGlobal;
    int pxSize;
  }optstore;

  //initialize optstore structure with initial values from config_and_coord.h
  optstore.steps = RUNS;
  optstore.steps2bmp = STEPS_2_BMP;
  optstore.resolution = RESOLUTION;
  optstore.specialBmpDir = false;
  optstore.popFactor = 0; 
  optstore.popSizeGlobal = INITIAL_POPULATION_SIZE;
  optstore.worldSizeGlobal = N_WORLD_SIZE;
  optstore.pxSize = 4;

  //parse options
  opt = getopt(argc, argv, optstring);
  while( opt != -1 ){
    switch(opt){
    case 'x':
      optstore.pxSize = atoi(optarg);
      break;
    case 's':
      optstore.steps = atoi(optarg);
      break;
    case  'b':
      optstore.steps2bmp = atoi(optarg);
      break;
    case  'r':
      optstore.resolution = atoi(optarg);
      break;
    case 'd':
      optstore.specialBmpDir = true;
      strcpy(optstore.bmpDir, optarg);
      break;
    case 'w':
      optstore.worldSizeGlobal = atoi(optarg);
      break;
    case 'p':
      optstore.popFactor = atoi(optarg);
      break;
    case 'h':
    case '?':
      printf("usage:\n\t-s [#steps]: steps of the simulation\n\t-b [#steps]: steps to .bmp file creation\n\t-r [#fields]: resolution of the bmp-files\n\t-d [directory name]: directory in ./bmps/ to save .bmp files\n\t-w [size]: Global world size\n\t-p [denominator]: Population size as a fraction of the world size\n");
    printf("********\n");
    return(EXIT_SUCCESS);
    }
    opt = getopt(argc, argv, optstring);
  }

  if(optstore.popFactor){
    optstore.popSizeGlobal = optstore.worldSizeGlobal*optstore.worldSizeGlobal / optstore.popFactor;
  }

  printf("****SOME PARAMETERS****\n");
  printf("Seriell Version!\nGlobal world-size N: %d\n",optstore.worldSizeGlobal);
  printf("Global start population: %d\n", optstore.popSizeGlobal);
  printf("Resolution: %d\nSteps :%d\nBmp printed each %d step\n", optstore.resolution, optstore.steps, optstore.steps2bmp);
  printf("***********************\n");


  srand(time(NULL));
 
  earth = genesis(insert_initial_random_population, optstore.popSizeGlobal, optstore.worldSizeGlobal, 0, 0,0);

  info = init_info(earth, optstore.resolution, 0, 0);
  size = earth->n / optstore.resolution;
  size = size * size;
  printf("size: %d\n",size);

  for(i=0; i<optstore.steps; i++){
    global_death(check_age_and_maybe_die, earth);
    die_by_overpopulation(earth);
    migration(earth);
    mating_and_birth(earth);
    if(i%optstore.steps2bmp == 0){
      printf("Month: %d\tFemales: %d\tMales: %d\n", i, earth->females.org_num, earth->males.org_num);
      get_info_alt(simple_autosomal_dominant_recessive_phenotype, earth, info, optstore.resolution);
      generation = i/100;
      if(optstore.specialBmpDir){
	draw_bmp_plus(info, size, generation, optstore.resolution, optstore.bmpDir, optstore.pxSize);
      }else{
	draw_bmp_plus(info, size, generation, optstore.resolution, "BmpSeriell", optstore.pxSize);
      }
    }
    if(earth->females.org_num == 0 && earth->males.org_num == 0){
      printf("Extinguished in month %d\n", i);
      break;
    }
  }

  free(info);
  armageddon(earth);

  return(EXIT_SUCCESS);
}
