

#include "org_and_env.h"

bool female(o_pointer org){
  if(org->sex_chromosomes[0] && org->sex_chromosomes[GENES_ON_GONOSOMES]){
    return(true);
  }else{
    return(false);
  }
}

int total_population_from_density(world*wrld){
  //debugging function
  int i, result=0;
  for(i=0; i<(wrld->n*wrld->n); i++){
    result+=wrld->population_density[i];
  }
  return(result);
}

int getID(world*wrld){
  org_id id;
  wrld->nextID++;
  id = wrld->nextID;
  id = id * 1000 + wrld->process;
  return(id);
}

void increase_density_at(world*wrld, int x, int y){
  wrld->population_density[wrld->n*(x-wrld->corner_x)+(y-wrld->corner_y)]++;
  return;
}

void decrease_density_at(world*wrld, int x, int y){
  wrld->population_density[wrld->n*(x-wrld->corner_x)+(y-wrld->corner_y)]--;
  return;
}

void decrease_density_from_coord_list(world*wrld, coordinate_list*list){
  int i;
  for(i=0; i<list->entries; i++){
    decrease_density_at(wrld, list->x[i], list->y[i]);
  }
  return;
}

void init_list(org_list*listing, int initial_size){
  listing->sorted = false;
  listing->org_num = 0;
  listing->size = initial_size;
  listing->list = malloc( initial_size*sizeof(o_pointer) );
  return;
}

org_list*new_list(int size){
  org_list*newList = malloc( sizeof(org_list) );
  init_list(newList, size);
  return(newList);
}

void delete_list(org_list*to_delete){
  free(to_delete->list);
  free(to_delete);
  return;
}

void list_resize(org_list*listing){
  listing->size*=2;
  listing->list = realloc( listing->list, listing->size*sizeof(o_pointer) );
  return;
}

void list_insert(org_list*listing, o_pointer to_insert){
  if(listing->size == listing->org_num) list_resize(listing);
  listing->list[listing->org_num] = to_insert;
  listing->org_num++;
  return;
}

o_pointer list_extract_organism(org_list*listing, int org_pos){
  o_pointer return_org = listing->list[org_pos];
  listing->org_num--;
  listing->list[org_pos] = listing->list[listing->org_num];
  return(return_org);
}

void list_kill_organism(org_list*listing, int org_pos){
  o_pointer to_kill = list_extract_organism(listing, org_pos);
  free(to_kill);
  return;
}

int cmp_pos(o_pointer org1, o_pointer org2){
  //returns 1 if the position of org1 is bigger
  if(org1->x > org2->x){
    return(1);
  }else if(org1->x == org2->x && org1->y > org2->y){
    return(1);
  }else{
    return(0);
  }
}
 
void list_sort(org_list*listing){
  int i, j, pos_max;
  o_pointer max;
  for(i = listing->org_num-1; i>0; i--){
    max = listing->list[0];
    pos_max = 0;
    for(j=1; j<=i; j++){
      if(cmp_pos(listing->list[j], max)){
	max = listing->list[j];
	pos_max = j;
      }
    }
    listing->list[pos_max] = listing->list[i];
    listing->list[i] = max;
  }
  listing->sorted = true;
  return;
}

int getPos(world*wrld, o_pointer org){
  int x,y,n;
  n=wrld->n;
  x=org->x;
  y=org->y;
  return(n*x+y);
}

bool relatives(o_pointer eva, o_pointer adam){
  if( adam->mother_id == 0 || eva->mother_id == 0 ) return(false);
  if( (adam->mother_id == eva->mother_id) || (adam->father_id == eva->father_id) ) return(true);
  if( (adam->mother_id == eva->id) || (eva->father_id == adam->id) ) return(true);
  return(false);
}


bool avoid_incest(o_pointer f, o_pointer m){
  if(m->age < ADULT_AGE_MALE) return(false);
  if(relatives(f,m)) return(false);
  return(true);
}

bool adult_partner(o_pointer f, o_pointer m){
  if(m->age < ADULT_AGE_MALE){
    return(false);
  }
  return(true);
}

bool coord_partner(o_pointer f, o_pointer m){
  if((f->x != m->x) || (f->y != m->y)) return(false);
  return(avoid_incest(f, m));
}

org_list*list_potential_partners(world*wrld, o_pointer female){
  //only coord_partner works here!!
  int i;
  bool found = false;
  org_list *partners = NULL;
  for(i=0; i < wrld->males.org_num; i++){
    if(coord_partner(female, wrld->males.list[i])){
      if(!found){
	partners = new_list(LIST_INIT_SIZE);
      }
      list_insert(partners, wrld->males.list[i]);
    }
  }
  return(partners);
}


o_pointer choose_a_male(org_list*candidates, o_pointer female){
  // idea choose male with smalles diffence in age, if same age, prefer closest
  o_pointer candidate;
  int i;
  if(candidates->org_num == 0) return(NULL);
  candidate = candidates->list[0];
  for(i=1; i<candidates->org_num; i++){
    if( fabs(candidates->list[i]->age - female->age) < fabs(candidate->age - female->age) ){
      candidate = candidates->list[i];
    }
  }
  return(candidate);
}

void fertilization(o_pointer mother, o_pointer father){
  if(father == NULL) return;
  mother->pregnant = 1;
  mother->month = 0;
  mother->partner_id = father->id;
  memcpy(mother->f_sex_chromosomes, father->sex_chromosomes, 2 * GENES_ON_GONOSOMES * sizeof(short int));
  memcpy(mother->f_chromosomes, father->chromosomes, CHROMOSOMES_DIPLOID * GENES_ON_AUTOSOMES * sizeof(short int));
  return;
}

void birth(world*wrld, o_pointer mother){
  o_pointer child;
  int gene;
  int mo_g, fa_g, mo_a[CHROMOSOMES_HAPLOID], fa_a[CHROMOSOMES_HAPLOID], g, c, children, ch;
  mother->pregnant = 0;  
  if( !check_phenotype(mother, MORE_CHILDREN) ){
    children = rand()%(MAX_CHILDREN_MUTANT - MIN_CHILDREN_MUTANT + 1) + MIN_CHILDREN_MUTANT;
  }else{
    children = rand()%(MAX_CHILDREN - MIN_CHILDREN + 1) + MIN_CHILDREN;
  }
  for(ch = 0; ch<children; ch++){
    increase_density_at(wrld, mother->x, mother->y);
    child = malloc(sizeof(organism));
    child->pregnant = 0;
    child->age = 0;
    child->x = mother->x;
    child->y = mother->y;
    child->mother_id = mother->id;
    child->father_id = mother->partner_id;
    child->id = getID(wrld);

    mo_g = rand()%2;
    fa_g = rand()%2;
    for(g=0; g<GENES_ON_GONOSOMES; g++){
      child->sex_chromosomes[0*GENES_ON_GONOSOMES+g] = mother->sex_chromosomes[mo_g*GENES_ON_GONOSOMES+g];
      child->sex_chromosomes[1*GENES_ON_GONOSOMES+g] = mother->f_sex_chromosomes[fa_g*GENES_ON_GONOSOMES+g];
    }

    for(c=0; c<CHROMOSOMES_HAPLOID; c++){
      mo_a[c] = rand()%2;
      fa_a[c] = rand()%2;
    }
    for(c=0; c<CHROMOSOMES_HAPLOID; c++){
      for(g=0; g<GENES_ON_AUTOSOMES; g++){
	gene = c * GENES_ON_AUTOSOMES + g;
	child->chromosomes[gene] = mother->chromosomes[gene + mo_a[c] * CHROMOSOMES_HAPLOID * GENES_ON_AUTOSOMES];
	child->chromosomes[gene + CHROMOSOMES_HAPLOID * GENES_ON_AUTOSOMES] = mother->f_chromosomes[gene + fa_a[c] * CHROMOSOMES_HAPLOID * GENES_ON_AUTOSOMES];
      }
    }
    if(female(child)){
      //female
      list_insert(&wrld->females, child);
    }else{
      //male
      list_insert(&wrld->males, child);
    }
  }
  return;
}

void mating_and_birth(world*wrld){
  int i;
  o_pointer mother, father;
  org_list*potential_partners;
  for(i=0; i<wrld->females.org_num; i++){
    mother = wrld->females.list[i];
    if(mother->pregnant){
      if(mother->month >= PREGNANCY_TIME){
	birth(wrld, mother);
      }
    }else{
      if(mother->age >= ADULT_AGE_FEMALE){
	potential_partners = list_potential_partners(wrld, mother);
	if(potential_partners != NULL){
	  father = choose_a_male(potential_partners, mother);
	  fertilization(mother, father);
	  delete_list(potential_partners);
	}
      }
    }
  }
  return;
}

bool check_age_and_maybe_die(world*wrld, org_list*listing, int org_pos){
  //returns true if the organism dies
  double x, die;
  o_pointer org = listing->list[org_pos];
  if( !check_phenotype(org, LONG_LIVE) ){
    x = (double) ( DEATH_CONST * (org->age - MAX_AGE_MUTANT) );
  }else{
    x = (double) ( DEATH_CONST * (org->age - MAX_AGE) );
  }
  x = exp(x);
  die = (double) rand() / RAND_MAX;
  if(die < x){ 
    decrease_density_at(wrld, org->x, org->y);
    list_kill_organism(listing, org_pos);
    return(true);
  }
  return(false);
}

void global_death( bool (*deathfunc)(world*wrld, org_list*listing, int org_pos), world*wrld){
  int i=0;
  bool someone_died;
  while(i < wrld->females.org_num){
    someone_died = deathfunc(wrld, &wrld->females, i);
    if(!someone_died) i++;
  }
  i = 0;
  while(i < wrld->males.org_num){
    someone_died = deathfunc(wrld, &wrld->males, i);
    if(!someone_died) i++;
  }
  return;
}

double death_parable(double population_density){
  double x;
  x = DEATH_RATE * pow((population_density-OPTIMAL_POPULATION_PER_FIELD)/(MAX_POPULATION_PER_FIELD-OPTIMAL_POPULATION_PER_FIELD), DEATH_COEFFICIENT);
  return(x);
}

double death_sigmoid(double population_density){
  double t, x;
  t = SIGMOID_SLOPE * (population_density - MAX_POPULATION_PER_FIELD);
  x = DEATH_RATE * 1.0/(1.0 + exp(-t));
  return(x);
}

bool overpopulation_death(world*wrld, double (*death_math)(double population_density), org_list*listing, int pos){
  o_pointer org = listing->list[pos];
  int position = (org->x - wrld->corner_x)*wrld->n+(org->y - wrld->corner_y);
  double die, x;
  double pd = (double) wrld->population_density[position];
  x = death_math(pd);
  die = (double) rand() / RAND_MAX;
  if(check_phenotype(org, BETTER_SURVIVE)){
    x = x * BETTER_SURVIVAL_FACTOR;
  }
  if(die<x){
    //    list_kill_organism(listing, pos);
    return(true);
  }
  return(false);
}
 
void die_by_overpopulation(world*wrld){
  //kill and collect coordinates in list, then adjust density-map!
  int i=0;
  bool dead;
  coordinate_list * coords = new_coord_list(MAX_POPULATION_PER_FIELD);
  while(i < wrld->males.org_num){
    dead = overpopulation_death(wrld, death_sigmoid, &wrld->males, i);
    if(dead){ 
      add_2_coord_list(coords, wrld->males.list[i]->x, wrld->males.list[i]->y);
      list_kill_organism(&(wrld->males), i);
    }else{
      i++;
    }
  }
  i=0;
  while(i < wrld->females.org_num){
    dead = overpopulation_death(wrld, death_sigmoid, &wrld->females, i);
    if(dead) {
      add_2_coord_list(coords, wrld->females.list[i]->x, wrld->females.list[i]->y);
      list_kill_organism(&(wrld->females), i);
    }else{
      i++;
    }
  }
  decrease_density_from_coord_list(wrld, coords);
  delete_coord_list(coords);
  return;
}

void insert_new_random_organism(world*wrld){
  o_pointer org;
  int c, g;
  double r;
  org = malloc(sizeof(organism));
  org->pregnant = 0;
  org->father_id = 0;
  org->mother_id = 0;
  org->month = 0;
  org->age = rand()%MAX_AGE;
  org->sex_chromosomes[0] = 1;
  org->sex_chromosomes[GENES_ON_GONOSOMES] = rand()%2;

  // new probabilities for recessive alleles here!!
  for(c=0; c<CHROMOSOMES_DIPLOID; c++){
    for(g=0; g<GENES_ON_AUTOSOMES; g++){
      r = (double) rand() / (double) RAND_MAX;
      if( r < PROB_REC ){
	org->chromosomes[c*GENES_ON_AUTOSOMES+g] = 0;
      }else{
	org->chromosomes[c*GENES_ON_AUTOSOMES+g] = 1;
      }
    }
  }

  org->x = (rand() % wrld->n) + wrld->corner_x;
  org->y = (rand() % wrld->n) + wrld->corner_y;
  org->id = getID(wrld);
  if(female(org)){
    list_insert(&wrld->females, org);
  }else{
    list_insert(&wrld->males, org);
  }
  increase_density_at(wrld, org->x, org->y);
  return;
}

void insert_initial_random_population(world*wrld, int popSize){
  int i;
  for(i=0; i<popSize; i++){
    insert_new_random_organism(wrld);
  }
  return;
}

world* genesis( void (*initialPopFunc)(world*wrld, int popSize), int sizeOfInitPop, int world_size_n, unsigned int corner_x, unsigned int corner_y, int process){
  world* thisPlace;
  thisPlace = malloc( sizeof(world) );
  thisPlace->process = (org_id) process;
  thisPlace->corner_x = corner_x;
  thisPlace->corner_y = corner_y;
  thisPlace->nextID = 0;
  thisPlace->n = world_size_n;
  thisPlace->population_density = calloc(thisPlace->n * thisPlace->n, sizeof(int));
  init_list(&thisPlace->males, sizeOfInitPop);
  init_list(&thisPlace->females, sizeOfInitPop);
  initialPopFunc(thisPlace, sizeOfInitPop);
  return(thisPlace);
}

void armageddon(world*wrld){
  //empty both lists of organisms
  int i;
  for(i=0; i < wrld->males.org_num; i++) free(wrld->males.list[i]);
  for(i=0; i < wrld->females.org_num; i++) free(wrld->females.list[i]);
  free( wrld->females.list );
  free( wrld->males.list );
  free( wrld->population_density );
  free( wrld );
  return;
}
 
void move(world*wrld, organism* org){
  int direction, new_x, new_y;
  org->age++;
  if(org->pregnant) org->month++;
  decrease_density_at(wrld, org->x, org->y);
  direction = rand() % 8;
  new_x = org->x;
  new_y = org->y;
  switch(direction){
  case 0: 
    new_x--;
    new_y--;
    break;
  case 1:
    new_x--;
    break;
  case 2:
    new_y--;
    break;
  case 3:
    new_x++;
    new_y--;
    break;
  case 4:
    new_x--;
    new_y++;
    break;
  case 5:
    new_x++;
    break;
  case 6:
    new_y++;
    break;
  case 7:
    new_x++;
    new_y++;
    break;
  }
  new_x = (new_x + wrld->n) % wrld->n;
  new_y = (new_y + wrld->n) % wrld->n;
  org->x = new_x;
  org->y = new_y;
  increase_density_at(wrld, org->x, org->y);
  return;
}

void migration(world*wrld){
  int i;
  wrld->month++;
  for(i=0; i<wrld->females.org_num; i++) move(wrld, wrld->females.list[i]);
  for(i=0; i<wrld->males.org_num; i++) move(wrld, wrld->males.list[i]);
  return;
}


bool check_population(world*wrld){
  int i, sum=0;
  for(i=0; i < wrld->n*wrld->n; i++){
    sum += wrld->population_density[i];
  }
  if(sum == wrld->females.org_num + wrld->males.org_num){
    return(true);
  }else{
    return(false);
  }
}
