
#ifndef ORG_AND_ENV_H
#define ORG_AND_ENV_H

#include "config_and_coord.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include "mpi.h"

#define MPI_ORG_ID MPI_UNSIGNED_LONG

typedef unsigned long int org_id;

typedef struct{
  // father information for children
  short int f_chromosomes[CHROMOSOMES_DIPLOID * GENES_ON_AUTOSOMES];
  short int f_sex_chromosomes[2 * GENES_ON_GONOSOMES];
  short int month;
  org_id partner_id;

  int pregnant; 

  short int sex_chromosomes[2 * GENES_ON_GONOSOMES];  //FIRST GENE: 1 is x, 0 is y
  short int chromosomes[CHROMOSOMES_DIPLOID * GENES_ON_AUTOSOMES]; //1 is dominant, 0 recessive
  int age;
  unsigned int x;
  unsigned int y;
  org_id id;
  org_id mother_id;
  org_id father_id;
}organism;

typedef organism* o_pointer;

typedef o_pointer* o_list;

typedef struct{
  o_list list;
  int size;
  int org_num;
  bool sorted; 
}org_list;

typedef struct{
  org_list males;
  org_list females;
  unsigned int n;
  unsigned int month;
  org_id nextID;
  int*population_density;
  int corner_x;
  int corner_y;
  org_id process;
}world;


void list_kill_organism(org_list*listing, int org_pos);

void list_insert(org_list*listing, o_pointer to_insert);

void insert_initial_random_population(world*wrld, int popSize);

bool check_age_and_maybe_die(world*wrld, org_list*listing, int org_pos);

void mating_and_birth(world*wrld);

void global_death( bool (*deathfunc)(world*wrld, org_list*listing, int org_pos), world*wrld);
 
void die_by_overpopulation(world*wrld);

world* genesis( void (*initialPopFunc)(world*wrld, int popSize), int sizeOfInitPop, int world_size_n, unsigned int corner_x, unsigned int corner_y, int process);

void armageddon(world*wrld);

void seeTheWorld(world*wrld);

void move(world*wrld, organism* org);
 
void migration(world*wrld);

bool female(o_pointer org);

bool check_population(world*wrld);

#endif
