 
#include <stdio.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#include "mpi.h"

#include "config_and_coord.h"
#include "org_and_env.h"
#include "parallel_org.h"
#include "phenotypes.h"

#define TYPES_ORGANISM 13

int main(int argc, char*argv[]){
  
  world*earth;
  int arrivals[8];
  o_pointer incoming_organisms[8];
  int success, i, run, processes, rank;
  int x, y, n, directions[8];
  int corner_x, corner_y;
  int boxes, generation = 0;
  bool first_sending = true;

  MPI_Datatype mpi_organism;
  MPI_Datatype types_organism[TYPES_ORGANISM] = {MPI_SHORT, MPI_SHORT, MPI_SHORT, MPI_ORG_ID, MPI_INT, MPI_SHORT, MPI_SHORT, MPI_INT, MPI_UNSIGNED, MPI_UNSIGNED, MPI_ORG_ID, MPI_ORG_ID, MPI_ORG_ID};
  int org_len[TYPES_ORGANISM] = {CHROMOSOMES_DIPLOID*GENES_ON_AUTOSOMES, 2*GENES_ON_GONOSOMES, 1, 1, 1, 2*GENES_ON_GONOSOMES, CHROMOSOMES_DIPLOID*GENES_ON_AUTOSOMES, 1, 1, 1, 1, 1, 1};
  MPI_Aint org_disp[TYPES_ORGANISM];
  MPI_Request request_send[8], request_receive[8];
  MPI_Status status_send[8], status_receive[8];

  MPI_Request *request_phenotypes;  //allocate memory, one for each environment process, many for reiceiver and printer
  MPI_Status *status_phenotypes;

  MPI_Datatype mpi_phenotypes;
  MPI_Datatype pheno_types[4] = {MPI_UNSIGNED, MPI_UNSIGNED, MPI_UNSIGNED, MPI_UNSIGNED};
  int phenotypes_len[4] = {1, 1, 1, NR_PHENOTYPES};
  MPI_Aint phenotypes_disp[4];

  train*zug = build_a_train();
  organism schuh;
  population_information phe, *phenotypes;

  int opt;
  static const char *optstring = "x:s:b:r:d:w:p:h?";

  struct{
    int steps;
    int steps2bmp;
    int resolution;
    bool specialBmpDir;
    char bmpDir[16];
    int worldSizeGlobal;
    int worldSizeLocal;
    int popFactor;
    int popSizeGlobal;
    int popSizeLocal;
    int pxSize;
  }optstore;

  //initialize optstore structure with initial values from config_and_coord.h
  optstore.steps = RUNS;
  optstore.steps2bmp = STEPS_2_BMP;
  optstore.resolution = RESOLUTION;
  optstore.specialBmpDir = false;
  optstore.popFactor = 0;
  optstore.popSizeGlobal = 0;
  optstore.popSizeLocal = INITIAL_POPULATION_SIZE;
  optstore.worldSizeGlobal = 0;
  optstore.worldSizeLocal = N_WORLD_SIZE;
  optstore.pxSize = 5;

  //MPI initialisation
  success = MPI_Init(&argc, &argv);

  if(success != MPI_SUCCESS){
    printf ("Error starting MPI. Terminating..\n");
    MPI_Abort(MPI_COMM_WORLD, success);
    return(EXIT_FAILURE);
  }

  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &processes);


  //parse options
  opt = getopt(argc, argv, optstring);
  while( opt != -1 ){
    switch(opt){
    case 'x':
      optstore.pxSize = atoi(optarg);
      break;
    case 's':
      optstore.steps = atoi(optarg);
      break;
    case  'b':
      optstore.steps2bmp = atoi(optarg);
      break;
    case  'r':
      optstore.resolution = atoi(optarg);
      break;
    case 'd':
      optstore.specialBmpDir = true;
      strcpy(optstore.bmpDir, optarg);
      break;
    case 'w':
      optstore.worldSizeGlobal = atoi(optarg);
      break;
    case 'p':
      optstore.popFactor = atoi(optarg);
      break;
    case 'h':
    case '?':
      if(rank==0){
	printf("usage:\n\t-s [#steps]: steps of the simulation\n\t-b [#steps]: steps to .bmp file creation\n\t-r [#fields]: resolution of the bmp-files\n\t-d [directory name]: directory in ./bmps/ to save .bmp files\n\t-w [size]: Global world size\n\t-p [denominator]: Population size as a fraction of the world size\n\t-x [pixel size]: pixelsize for each position in bmp printing\n");
	printf("********\n");
      }
    MPI_Finalize();
    return(EXIT_SUCCESS);
    }
    opt = getopt(argc, argv, optstring);
  }


  //init random-machine
  srand(time(NULL)+rank);


  if(processes < 5){
    printf ("This needs at least five processes! Terminating.. \n");
    MPI_Finalize();
    return(EXIT_FAILURE);
  }
 
  
  //coordinates and neighbors
  n = floor(sqrt(processes-1));
  x = rank / n;
  y = rank % n;

  //initilize worldsize and size of population if changed by option parsing
  if(optstore.worldSizeGlobal){
    optstore.worldSizeLocal = optstore.worldSizeGlobal / n;
    if(optstore.worldSizeGlobal % n != 0){
      printf("WARNING: Global worldsize does not fit Process number,\nfloor function is used to determine local worldsiz!\n");
    }
  }else{
    optstore.worldSizeGlobal = n*optstore.worldSizeLocal;
  }


  if(optstore.popFactor){    
    optstore.popSizeGlobal = optstore.worldSizeGlobal*optstore.worldSizeGlobal/optstore.popFactor;
    optstore.popSizeLocal = optstore.popSizeGlobal/(n*n);
    if(rank < optstore.popSizeGlobal % (n*n)) optstore.popSizeLocal++;
  }else{
    optstore.popSizeGlobal = optstore.popSizeLocal * n * n;
  }
  
  //print parameters and information about this simulation run
  if(rank == n*n){
    printf("****SOME PARAMETERS****\n");
    printf("Processes used: %d\nProcess N: %d\nGlobal world-size N: %d\nLocal size per process: %d\n", n*n+1, n, optstore.worldSizeGlobal, optstore.worldSizeLocal);
    printf("Global start population: %d\nLocal start population per process: %d\n", optstore.popSizeGlobal, optstore.popSizeLocal);
    printf("Resolution: %d\nSteps :%d\nBmp printed each %d step\n", optstore.resolution, optstore.steps, optstore.steps2bmp);
    printf("Pixel size: %d\n", optstore.pxSize);
    printf("***********************\n");
  }

  //calculate rank of processes for neighboring fields

  directions[0] = n * ((x-1+n)%n) + ((y-1+n)%n);
  directions[1] = n * x + ((y-1+n)%n);
  directions[2] = n * ((x+1)%n) + ((y-1+n)%n);
  directions[3] = n * ((x+1)%n) + y;
  directions[4] = n * ((x+1)%n) + ((y+1)%n);
  directions[5] = n * x + ((y+1)%n);
  directions[6] = n * ((x-1+n)%n) + ((y+1)%n);
  directions[7] = n * ((x-1+n)%n) + y;


  /*
   * Calculate dispositions and commit MPI-structures
   */

  //dispositions
  org_disp[0] = (void*) &schuh.f_chromosomes - (void*) &schuh;
  org_disp[1] = (void*) &schuh.f_sex_chromosomes - (void*) &schuh;
  org_disp[2] = (void*) &schuh.month - (void*) &schuh;
  org_disp[3] = (void*) &schuh.partner_id - (void*) &schuh;
  org_disp[4] = (void*) &schuh.pregnant - (void*) &schuh;
  org_disp[5] = (void*) &schuh.sex_chromosomes - (void*) &schuh;
  org_disp[6] = (void*) &schuh.chromosomes - (void*) &schuh;
  org_disp[7] = (void*) &schuh.age - (void*) &schuh;
  org_disp[8] = (void*) &schuh.x - (void*) &schuh;
  org_disp[9] = (void*) &schuh.y - (void*) &schuh;
  org_disp[10] = (void*) &schuh.id - (void*) &schuh;
  org_disp[11] = (void*) &schuh.mother_id - (void*) &schuh;
  org_disp[12] = (void*) &schuh.father_id - (void*) &schuh;


  phenotypes_disp[0] = (void*) &phe.x - (void*) &phe;
  phenotypes_disp[1] = (void*) &phe.y - (void*) &phe;
  phenotypes_disp[2] = (void*) &phe.total_nr_org - (void*) &phe;
  phenotypes_disp[3] = (void*) &phe.phenotypes - (void*) &phe;
    
  // commit datattypes
  MPI_Type_create_struct(TYPES_ORGANISM, org_len, org_disp, types_organism, &mpi_organism);
  MPI_Type_commit(&mpi_organism);

  MPI_Type_create_struct(4, phenotypes_len, phenotypes_disp, pheno_types, &mpi_phenotypes);
  MPI_Type_commit(&mpi_phenotypes);

  //calculate corner x and y and how many processes involved 
  corner_x = x * optstore.worldSizeLocal;
  corner_y = y * optstore.worldSizeLocal;


  boxes = pow((optstore.worldSizeLocal/optstore.resolution),2);
  

  if(rank < n*n){
    earth = genesis(insert_initial_random_population, optstore.popSizeLocal, optstore.worldSizeLocal, corner_x, corner_y, rank);
    phenotypes = init_info(earth, optstore.resolution, x, y);
    if(phenotypes == NULL){
      printf("Resolution does not fit the environment\n");
      MPI_Abort(MPI_COMM_WORLD, success);
      return(EXIT_FAILURE);
    }
    request_phenotypes = malloc(sizeof(MPI_Request));
    status_phenotypes = malloc(sizeof(MPI_Status));
  }else if(rank == n*n){
    phenotypes = malloc( boxes  * pow(n,2) * sizeof(population_information) ); 
    request_phenotypes = malloc(n*n*sizeof(MPI_Request));
    status_phenotypes = malloc(n*n*sizeof(MPI_Status));

  }
  

  for(run=0; run < optstore.steps; run++){

    // gather info every x-steps
        
    if(run % optstore.steps2bmp == 0){
      if(rank < n*n){
	if(first_sending){
	  first_sending = false;
	}else{
	  MPI_Wait(request_phenotypes, status_phenotypes);
	}
	get_info_alt(autosomal_phenotypes, earth, phenotypes, optstore.resolution);
	MPI_Isend(phenotypes, boxes, mpi_phenotypes, n*n, 0, MPI_COMM_WORLD, request_phenotypes); //fill in send info
      }else if(rank == n*n){
	printf("rank %d run %d\n",rank, run);
	for(i=0; i<n*n; i++){
	  MPI_Irecv( &(phenotypes[i*boxes]), boxes, mpi_phenotypes, i, 0, MPI_COMM_WORLD, &(request_phenotypes[i]) );
	}
	for(i=0; i<n*n; i++){
	  MPI_Wait(&(request_phenotypes[i]), &(status_phenotypes[i]));
	}
	if(optstore.specialBmpDir){
	  draw_bmp_plus(phenotypes, boxes*n*n, generation, optstore.resolution, optstore.bmpDir, optstore.pxSize);
	}else{
	  draw_bmp_plus(phenotypes, boxes*n*n, generation, optstore.resolution, NULL, optstore.pxSize);
	}
	generation++;
      }
    }
        

    if(rank < n*n){

      //check death, mating, birth and move
      mating_and_birth(earth);
      global_death(check_age_and_maybe_die, earth);
      die_by_overpopulation(earth);
      move_all_parallel( earth, zug,  n*optstore.worldSizeLocal);

      //send and receive how many organisms are coming 
      for(i=0; i<8; i++){
	MPI_Isend(&(zug->nr_passengers[i]), 1, MPI_INT, directions[i], 0, MPI_COMM_WORLD, &(request_send[i]));
	MPI_Irecv(&(arrivals[i]), 1, MPI_INT, directions[i], 0, MPI_COMM_WORLD, &(request_receive[i]));
      }
      for(i=0; i<8; i++){
	//wait for message passing to be done
	MPI_Wait(&(request_send[i]), &(status_send[i]));
	MPI_Wait(&(request_receive[i]), &(status_receive[i]));
      }
     
      //send and receive organisms
      for(i=0; i<8; i++){
	MPI_Isend(zug->passengers[i], zug->nr_passengers[i], mpi_organism, directions[i], 0, MPI_COMM_WORLD, &(request_send[i]));
	incoming_organisms[i] = malloc(arrivals[i]*sizeof(organism));
	MPI_Irecv(incoming_organisms[i], arrivals[i], mpi_organism, directions[i], 0, MPI_COMM_WORLD, &(request_receive[i]));
      }
      
      for(i=0; i<8; i++){
	//wait and check status
	MPI_Wait(&(request_send[i]), &(status_send[i]));
	MPI_Wait(&(request_receive[i]), &(status_receive[i]));
      }
      
      //place organisms into environment!!
      for(i=0; i<8; i++){
	set_free(earth, incoming_organisms[i], arrivals[i]);
      }
      
    }
  }
  train_destruction(zug);
  if(rank < n*n){
    printf("rank %d: %d organisms\n", rank, earth->females.org_num + earth->males.org_num);
    armageddon(earth);
  }
  
  if(rank <= n*n){
    free(status_phenotypes);
    free(request_phenotypes);
    free(phenotypes);
  }
  
  //end mpi
  MPI_Type_free(&mpi_organism);
  MPI_Type_free(&mpi_phenotypes);

  MPI_Finalize();
  return(EXIT_SUCCESS);
  
}
