
#include "phenotypes.h" 

int simple_autosomal_dominant_recessive_phenotype(o_pointer org){
  if( !(org->chromosomes[0] || org->chromosomes[GENES_ON_AUTOSOMES]) ){
    return(0);
  }else{
    return(1);
  }
}

int autosomal_phenotypes(o_pointer org){
  int chromosome, g, gene, phenotype=0, type;
  int genes = CHROMOSOMES_HAPLOID * GENES_ON_AUTOSOMES;
  for(chromosome = 0; chromosome < CHROMOSOMES_HAPLOID; chromosome++){
    for(g = 0; g < GENES_ON_AUTOSOMES; g++){
      gene = chromosome * GENES_ON_AUTOSOMES + g;
      type = org->chromosomes[gene] || org->chromosomes[gene + CHROMOSOMES_HAPLOID * GENES_ON_AUTOSOMES]; //1 or 0
      phenotype += type * pow(2, gene); 
    }
  }
  return(phenotype);
}

bool check_phenotype(o_pointer org, int gene_nr){
  int type = autosomal_phenotypes(org);
  int bit = type / pow(2, gene_nr);
  bit = bit % 2; 
  if(bit){
    return(true);
  }else{
    return(false);
  }
}

bool check_gene_phenotype(int type_nr, int gene_nr){
  int bit = type_nr / pow(2, gene_nr);
  bit = bit % 2; 
  if(bit){
    return(true);
  }else{
    return(false);
  }
}

population_information* init_info(world* wrld, int field_size, int process_x, int process_y){
  int fields_n, x, y;
  population_information* info;
  if(wrld->n % field_size != 0) return(NULL);
  fields_n = wrld->n / field_size;
  info = malloc( fields_n*fields_n*sizeof(population_information) );
  for(x=0; x<fields_n; x++){
    for(y=0; y<fields_n; y++){
      info[x * fields_n + y].x = x + (process_x*fields_n);
      info[x * fields_n + y].y = y + (process_y*fields_n);
    }
  }
  return(info);
}


void get_info_alt(int(*phenotype)(o_pointer org), world* wrld, population_information*info, int field_size){
  int i,j, fields_n, x, y;
  int from, to, big_x, big_y;

  fields_n = wrld->n / field_size;
  for(i=0; i<fields_n*fields_n; i++){
    info[i].total_nr_org = 0;
    for(j=0; j<NR_PHENOTYPES; j++){
      info[i].phenotypes[j] = 0;
    }
  }
  for(x=0; x<fields_n; x++){
    for(y=0; y<fields_n; y++){
      for(i=0; i<field_size; i++){
	for(j=0; j<field_size; j++){
	  big_x = x * field_size + i;
	  big_y = y * field_size + j;
	  info[x * fields_n + y].total_nr_org += wrld->population_density[big_x * wrld->n + big_y];
	}
      }
    }
  }

  for(j=0; j<2; j++){
    org_list*organisms;
    if(j){
      organisms = &(wrld->females);
    }else{
      organisms = &(wrld->males);
    }
    for(i=0; i < organisms->org_num; i++){
      o_pointer this_org = organisms->list[i];
      int type = phenotype(this_org);
      x = (this_org->x - wrld->corner_x) / field_size;
      y = (this_org->y - wrld->corner_y) / field_size;
      info[x * fields_n + y].phenotypes[type]++;
    }
  }
  return;
}



double minimum(double a, double b){
  if(a<b) {
    return(a);
  }else{
    return(b);
  }
}


void set_pixel_with_size(int px_size, bmpfile_t *bmpfile, uint32_t x, uint32_t y, rgb_pixel_t pixel){
  int i, j, startX=px_size*x, startY=px_size*y;
  for(i=0; i<px_size; i++){
    for(j=0; j<px_size; j++){
      bmp_set_pixel(bmpfile, startX+i, startY+j, pixel);
    }
  }
  return;
}



void draw_bmp_plus(population_information*info, int size, int generation, int field_size, char*directory_name, int pixelsize){

  bmpfile_t *bmp;
  char filename[255], pathname[255], dir[16], system_call[255];
  int i, n, type, gene_nr;
  double range_dr = 1;
  double step_dr = range_dr / 8;
  rgb_pixel_t colors[8] = {{255,255,255,0},{50,205,50,0}, {34,139,34,0}, {252, 215, 0, 0}, {255, 140, 0, 0}, {255, 0, 0, 0}, {199, 0, 0, 0},{0,0,0,0}};
  rgb_pixel_t colors_dr[8] = {{255,255,255,0},{50,205,50,0}, {34,139,34,0}, {252, 215, 0, 0}, {255, 140, 0, 0}, {255, 0, 0, 0}, {199, 0, 0, 0},{0,0,0,0}};
  rgb_pixel_t seperation = {139,137,137};
  n = (int) sqrt(size);

  if(directory_name == NULL){
    snprintf(dir, (size_t)16, "");
  }else{
    snprintf(dir, (size_t)16, "/%s", directory_name);
  }
  snprintf(pathname, (size_t)255, "/home/petersen/ParallelPopulationDynamics/bmps%s/", dir);
  snprintf(filename, (size_t)255, "%s/bmp_%d", pathname, generation);
  snprintf(system_call, (size_t)255, "chmod 700 %s", pathname );

  mkdir(pathname);
  
  system(system_call);

  bmp = bmp_create((n*2+1)*pixelsize, (n*2+1)*pixelsize, 24);

  //simple population density

  for(i=0; i<size; i++){
    rgb_pixel_t pix, pix_2;
    double density = (double) info[i].total_nr_org/( (double) (field_size*field_size) * MAX_POPULATION_PER_FIELD);
    if(density < 0.01){
      memcpy(&pix, &colors[0], sizeof(rgb_pixel_t));
    }else if(density < 0.125){
      memcpy(&pix, &colors[1], sizeof(rgb_pixel_t));
    }else if(density < 0.25){
      memcpy(&pix, &colors[2], sizeof(rgb_pixel_t));
    }else if(density < 0.375){
      memcpy(&pix, &colors[3], sizeof(rgb_pixel_t));
    }else if(density < 0.5){
      memcpy(&pix, &colors[4], sizeof(rgb_pixel_t));
    }else if(density < 0.625){
      memcpy(&pix, &colors[5], sizeof(rgb_pixel_t));
    }else if(density < 0.75){
      memcpy(&pix, &colors[6], sizeof(rgb_pixel_t));
    }else{
      memcpy(&pix, &colors[7], sizeof(rgb_pixel_t));
    }
    set_pixel_with_size(pixelsize, bmp, info[i].x, info[i].y, pix);
     

    for(gene_nr = 0; gene_nr < 3; gene_nr++){
    //visualize recessive
      double recessives = 0.0;
      for(type = 0; type < NR_PHENOTYPES; type++){
	if(!check_gene_phenotype(type, gene_nr)){
	  recessives += (double) info[i].phenotypes[type] / ( (double) (field_size*field_size) * MAX_POPULATION_PER_FIELD);
	}
      }
      if(recessives < 0.01){
	memcpy(&pix_2, &colors_dr[0], sizeof(rgb_pixel_t));
      }else if(recessives < step_dr){
	memcpy(&pix_2, &colors_dr[1], sizeof(rgb_pixel_t));
      }else if(recessives < step_dr * 2){
	memcpy(&pix_2, &colors_dr[2], sizeof(rgb_pixel_t));
      }else if(recessives < step_dr * 3){
	memcpy(&pix_2, &colors_dr[3], sizeof(rgb_pixel_t));
      }else if(recessives < step_dr * 4){
	memcpy(&pix_2, &colors_dr[4], sizeof(rgb_pixel_t));
      }else if(recessives < step_dr * 5){
	memcpy(&pix_2, &colors_dr[5], sizeof(rgb_pixel_t));
      }else if(recessives < step_dr * 6){
	memcpy(&pix_2, &colors_dr[6], sizeof(rgb_pixel_t));
      }else{
	memcpy(&pix_2, &colors_dr[7], sizeof(rgb_pixel_t));
      }
      switch(gene_nr){
      case 0:
	set_pixel_with_size(pixelsize, bmp, info[i].x + n + 1, info[i].y, pix_2);
	break;
      case 1:
	set_pixel_with_size(pixelsize, bmp, info[i].x, info[i].y + n + 1, pix_2);
	break;
      case 2:
	set_pixel_with_size(pixelsize, bmp, info[i].x + n + 1, info[i].y + n + 1, pix_2);
	break;	
      }
    }
  }

  for(i=0; i<2*n+1; i++){
    set_pixel_with_size(pixelsize, bmp, n, i, seperation);
    set_pixel_with_size(pixelsize, bmp, i, n, seperation);    
  }


  bmp_save(bmp, filename);
  bmp_destroy(bmp);

  return;
}

